"""
Interface objects: dict representation, directory and properties inodes.
"""

import py9p
import os

from iproute2fs.vfs import Inode

class InterfaceInode(Inode):
    """
    Interface directory.
    """
    def __init__(self, name, parent):
        Inode.__init__(self, name, parent, qtype=py9p.DMDIR)
        self.interface = parent.objects['links-by-name'][name]
        self.child_map = {
            "addresses":    AdressesInode,
            "flags":        FlagsInode,
            "mtu":          MtuInode,
            "neighbors":    NeighborsInode,
            "hwaddr":       HwAddressInode,
        }

class MtuInode(Inode):
    """
    Contains MTU, R/O
    """
    def sync(self):
        self.data.seek(0, os.SEEK_SET)
        self.data.truncate()
        self.data.write(str(self.parent.interface['mtu']))
        self.data.write("\n")

class FlagsInode(Inode):
    """
    Interface flags can be used to detect up/down interfaces,
    'cause interface's state not always returns relevant values.

    R/O
    """
    def sync(self):
        self.data.seek(0, os.SEEK_SET)
        self.data.truncate()
        self.data.write(",".join(self.parent.interface['flags']))
        self.data.write("\n")

class HwAddressInode(Inode):
    """
    MAC address, R/O
    """
    def sync(self):
        self.data.seek(0, os.SEEK_SET)
        self.data.truncate()
        self.data.write(self.parent.interface['hwaddr'])

class NeighborsInode(Inode):
    """
    ARP cache
    """
    def sync(self):
        self.data.seek(0, os.SEEK_SET)
        self.data.truncate()
        self.data.write("".join(
            [ "%15s%20s\n" % (x['dest'],x['lladdr'])
                for x in self.parent.interface['neighbors'].values() ]
            ))

class AdressesInode(Inode):
    """
    IPv4,6 addresses, R/W
    """
    def sync(self):
        self.addresses = []
        for x in self.parent.interface['addresses'].values():
            if x.has_key('local'):
                if x['local'] != x['address']:
                    self.addresses.append("%s/%s peer %s" %
                        (x['local'], x['mask'], x['address']))
                    continue
            self.addresses.append("%s/%s" % (x['address'], x['mask']))
        self.data.seek(0, os.SEEK_SET)
        self.data.truncate()
        self.data.write("".join([ "%s\n" % (x) for x in self.addresses ]))

    def commit(self):
        # get addr. list
        self.data.seek(0, os.SEEK_SET)
        chs = set(self.addresses)
        prs = set([ x.strip() for x in self.data.readlines() ])
        to_delete = chs - prs
        to_create = prs - chs
        try:
            [ self.storage.iproute2.del_addr(self.parent.interface['dev'],
                x.split()[0])
                for x in to_delete ]
            [ self.storage.iproute2.add_addr(self.parent.interface['dev'],
                x.split()[0])
                for x in to_create ]
        except Exception, e:
            print e

