"""
Playback engine for iproute2fs: routines to parse and handle
event records.
"""

from cxnet.netlink.rtnl import RtNetlinkEvent

class Interface(dict):
    """
    Extend interface's event to meet our requirements
    """
    def __init__(self, rt_dict):
        dict.__init__(self, rt_dict)
        self['addresses'] = {}
        self['neighbors'] = {}

class RoutingTable(dict):
    """
    Routing table
    """
    pass

class Route(RtNetlinkEvent):
    """
    A route
    """
    def __init__(self,event,interfaces):
        RtNetlinkEvent.__init__(self,event)
        self.interfaces = interfaces

    def __str__(self):
        s = ""
        if self.has_key('dst_prefix') and self.has_key('dst_len'):
            s += "%s/%s " % (self['dst_prefix'],self['dst_len'])
        else:
            s += "default "
        if self.has_key('gateway'):
            s += "via %s " % (self['gateway'])
        if self.has_key('prefsrc'):
            s += "source %s " % (self['prefsrc'])
        if self.has_key('output_link'):
            s += "dev %s " % (self.interfaces[self['output_link']]['dev'])
        if self.has_key('priority'):
            s += "priority %s " % (self['priority'])
        s += "\n"
        return s

@vars
class SyncMap (object):
    """
    Map to handle incoming events. Contains several maps,
    each specific to it's object type. Submaps are keyed
    by action types.
    """

    @vars
    class link:
        """
        Link, or interface, events.
        """

        def add(event, objects):
            """
            Add or update interface record after the event.
            """
            if not objects['links'].has_key(event['index']):
                # create new interface record
                objects['links-by-name'][event['dev']] = \
                    objects['links'][event['index']] = \
                    Interface(event)
            else:
                # update existing interface record
                [ objects['links'][event['index']].__setitem__(x, y)
                    for x, y in event.items() ]

        def remove(event, objects):
            """
            Remove interface record.
            """
            try:
                del objects['links'][event['index']]
                del objects['links-by-name'][event['dev']]
            except:
                pass

    @vars
    class address:
        """
        Address events. IPv4 as well as IPv6 addresses are supported.
        """

        def add(event, objects):
            """
            New address added
            """
            if event.has_key('local'):
                key = '%s/%s' % (event['local'], event['mask'])
            else:
                key = '%s/%s' % (event['address'], event['mask'])
            objects['links'][event['index']]['addresses'][key] = event

        def remove(event, objects):
            """
            Address is removed.
            """
            if event.has_key('local'):
                key = '%s/%s' % (event['local'], event['mask'])
            else:
                key = '%s/%s' % (event['address'], event['mask'])
            try:
                del objects['links'][event['index']]['addresses'][key]
            except:
                pass

    @vars
    class neigh:
        """
        ARP cache events
        """

        def add(event, objects):
            """
            ARP record added.
            """
            objects['links'][event['index']]['neighbors'][event['lladdr']] = event

        def remove(event, objects):
            """
            ARP record removed.
            """
            try:
                del objects['links'][event['index']]['neighbors'][event['lladdr']]
            except:
                pass

    @vars
    class route:
        """
        Routing events
        """

        def add(event, objects):
            """
            New route.
            """
            if not objects['routing'].has_key(event['t']):
                objects['routing'][event['t']] = RoutingTable()

            objects['routing'][event['t']][hash(event)] = Route(event,objects['links'])

        def remove(event, objects):
            """
            Route is removed
            """
            try:
                del objects['routing'][event['t']][event]
            except:
                pass

def sync(objects, iproute2, blocking=False):
    """
    Event handling loop, either blocking or not. Not blocking
    sync can be used to handle only already received events and
    then return. Blocking sync can be used as a thread to
    continuously update object records.
    """
    while True:
        events = iproute2.get(0, blocking)
        if len(events) == 0:
            break
        for event in events:
            SyncMap[event['type']][event['action']](event, objects)
